use std::{collections::BTreeMap, os::unix::fs::PermissionsExt, path::PathBuf};

use anyhow::Result;
use async_compression::{Level, tokio::bufread::ZstdEncoder};
use aws_sdk_s3::{
    self as s3,
    types::{CompletedMultipartUpload, CompletedPart},
};
use bytes::{BufMut, Bytes, BytesMut};
use futures::{future::join_all, stream::TryStreamExt};
use nix_compat::{
    nar::writer::r#async as nar,
    narinfo::{self, NarInfo, SigningKey},
    nixbase32,
    store_path::StorePath,
};
use sha2::{Digest, Sha256};
use tokio::{
    fs::{File, read_dir, read_link},
    io::{AsyncRead, BufReader},
    pin,
};
use tokio_stream::wrappers::ReadDirStream;
use tokio_util::io::{InspectReader, read_buf};
use tracing::debug;

use crate::{bindings::AsyncWriteAdapter, path_info::PathInfo, store::Store};

const MULTIPART_CUTOFF: usize = 1024 * 1024 * 5;

pub struct Uploader<'a> {
    signing_key: &'a SigningKey<ed25519_dalek::SigningKey>,
    path: PathInfo,
    s3_client: &'a s3::Client,
    bucket: String,
    store: &'a Store,
}

impl<'a> Uploader<'a> {
    pub fn new(
        signing_key: &'a SigningKey<ed25519_dalek::SigningKey>,
        path: PathInfo,
        s3_client: &'a s3::Client,
        bucket: String,
        store: &'a Store,
    ) -> Result<Self> {
        Ok(Self {
            signing_key,
            path,
            s3_client,
            bucket,
            store,
        })
    }

    pub async fn upload(&self) -> Result<()> {
        let mut nar_temp = File::open(tempfile::Builder::new().tempfile()?.path()).await?;
        self.make_nar(&mut nar_temp).await;

        // this goes to the .narinfo file
        let mut nar_hasher = Sha256::new();
        // this is the URL for file .narinfo points to
        let mut file_hasher = Sha256::new();
        let nar_reader = compress_and_hash_nar(nar_temp, &mut nar_hasher, &mut file_hasher);

        let buf = BytesMut::with_capacity(MULTIPART_CUTOFF);
        let 

        if first_chunk.len() < MULTIPART_CUTOFF {
            let put_object = self
                .s3_client
                .put_object()
                .bucket(&self.bucket)
                .key(&nar_url)
                .body(first_chunk.into())
                .send()
                .await?;
            debug!("put object: {:#?}", put_object);
        } else {
            let multipart = self
                .s3_client
                .create_multipart_upload()
                .bucket(&self.bucket)
                .key(&nar_url)
                .send()
                .await?;
            let upload_id = multipart.upload_id().unwrap();

            let mut parts = Vec::with_capacity(nar.len() / MULTIPART_CUTOFF);
            let chunks = nar.chunks(MULTIPART_CUTOFF);
            for (i, chunk) in chunks.enumerate() {
                parts.push(tokio::task::spawn(
                    self.s3_client
                        .upload_part()
                        .bucket(&self.bucket)
                        .key(&nar_url)
                        .upload_id(upload_id)
                        .part_number(i as i32 + 1)
                        .body(chunk.to_vec().into())
                        .send(),
                ));
            }

            let completed_parts = join_all(parts)
                .await
                .into_iter()
                .flatten()
                .collect::<Result<Vec<_>, _>>()?
                .into_iter()
                .enumerate()
                .map(|(i, part)| {
                    CompletedPart::builder()
                        .set_e_tag(part.e_tag().map(ToString::to_string))
                        .set_part_number(Some(i as i32 + 1))
                        .set_checksum_sha256(part.checksum_sha256().map(ToString::to_string))
                        .build()
                })
                .collect::<Vec<_>>();

            let completed_mp_upload = CompletedMultipartUpload::builder()
                .set_parts(Some(completed_parts))
                .build();

            let complete_mp_upload = self
                .s3_client
                .complete_multipart_upload()
                .bucket(&self.bucket)
                .key(&nar_url)
                .upload_id(upload_id)
                .multipart_upload(completed_mp_upload)
                .send()
                .await?;

            debug!("complete multipart upload: {:#?}", complete_mp_upload);
        }

        let narinfo_url = format!("{}.narinfo", self.path.digest());
        debug!("uploading narinfo with key {narinfo_url}");
        self.s3_client
            .put_object()
            .bucket(&self.bucket)
            .key(narinfo_url)
            .body(nar_info.to_string().as_bytes().to_vec().into())
            .send()
            .await?;
        debug!("done uploading narinfo");

        Ok(())
    }

    fn narinfo_from_nar(&self, nar: &[u8]) -> Result<NarInfo> {
        let mut hasher = Sha256::new();
        hasher.update(nar);
        let nar_hash: [u8; 32] = hasher.finalize().into();
        let mut nar_info = NarInfo {
            flags: narinfo::Flags::empty(),
            store_path: self.path.path.as_ref(),
            nar_hash,
            nar_size: nar.len() as u64,
            references: self.path.references.iter().map(StorePath::as_ref).collect(),
            signatures: Vec::new(),
            ca: None,
            system: None,
            deriver: None,
            compression: Some("zstd"),
            file_hash: None,
            file_size: None,
            url: "",
        };
        // signature consists of: store_path, nar_hash, nar_size, and references
        nar_info.add_signature(self.signing_key);
        Ok(nar_info)
    }

    async fn make_nar(&self, sink: &mut File) -> Result<()> {
        let nar = nar::open(sink).await?;
        let path = self.path.absolute_path();
        let metadata = File::open(&path).await?.metadata().await?;

        if metadata.is_symlink() {
            let target = read_link(&path).await?;
            nar.symlink(target.as_os_str().as_encoded_bytes()).await;
        } else if metadata.is_dir() {
            let mut nar = nar.directory().await?;
            nar_from_dir(path.into(), &mut nar).await;
            nar.close().await;
        } else if metadata.is_file() {
            let perms = metadata.permissions().mode();
            let mut executable = false;
            if (perms & 0o700) == 0o700 {
                executable = true;
            }

            let mut file = BufReader::new(File::open(&path).await?);
            nar.file(executable, metadata.len(), &mut file).await;
        }

        Ok(())
    }
}

async fn nar_from_dir(path: PathBuf, node: &mut nar::Directory<'_, '_>) -> Result<()> {
    let root = ReadDirStream::new(read_dir(&path).await?);
    let entries = root
        .map_ok(|x| (x.file_name(), x))
        .try_collect::<BTreeMap<_, _>>()
        .await?;

    // directory entries must be written in ascending order of name
    for (name, entry) in entries.iter() {
        let node = node.entry(name.as_encoded_bytes()).await?;
        let metadata = entry.metadata().await?;

        if metadata.is_symlink() {
            let target = read_link(entry.path()).await?;
            node.symlink(target.as_os_str().as_encoded_bytes()).await;
        } else if metadata.is_dir() {
            let mut node = node.directory().await?;
            Box::pin(nar_from_dir(entry.path(), &mut node)).await;
            node.close().await;
        } else if metadata.is_file() {
            let perms = metadata.permissions().mode();
            let mut executable = false;
            if (perms & 0o700) == 0o700 {
                executable = true;
            }

            let mut file = BufReader::new(File::open(entry.path()).await?);
            node.file(executable, metadata.len(), &mut file).await;
        }
    }
    Ok(())
}

fn compress_and_hash_nar(
    nar_file: File,
    nar_hasher: &mut Sha256,
    compressed_nar_hasher: &mut Sha256,
) -> impl AsyncRead {
    let nar_reader = InspectReader::new(nar_file, |x| nar_hasher.update(x));
    let nar_buf_reader = BufReader::new(nar_reader);

    let compressed_nar_reader = ZstdEncoder::with_quality(nar_buf_reader, Level::Default);
    InspectReader::new(compressed_nar_reader, |x| compressed_nar_hasher.update(x))
}

fn nar_url(compressed_nar_hash: &[u8]) -> String {
    format!("nar/{}.nar.zst", nixbase32::encode(compressed_nar_hash))
}

async fn read_buf_nar<S: AsyncRead + Unpin>(stream: &mut S, mut buf: BytesMut) -> Result<Bytes> {
    while buf.len() < buf.capacity() {
        let n = read_buf(stream, &mut buf).await?;
        
        if n == 0 {
            break;
        }
    }
    Ok(buf.freeze())
}
